#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "head.h"

#define U_LOW	-0.3
#define U_HIGH	1.0
#define V_LOW	0.0
#define V_HIGH	0.15

#define LOA	255
/*
#define THT	5
*/

#define DO_MK_CUV	0
#define DO_SOLVE_U	1
#define DO_SOLVE_V	2

unsigned char d2uc( d, l, h )
	double d, l, h ;
{
	if ( d >= h )
		return (unsigned char)255 ;
	else if ( d <= l )
		return (unsigned char)0 ;
	else
		return (unsigned char)(255.*(d-l)/(h-l)) ;
}

double fuv( u, v, a )
	double u, v, a ;
{
	return (u*(u-a)*(1.-u)-v) ;
}

double guv( u, v, b )
	double u, v, b ;
{
	return (u-b*v) ;
}

int ip4loa( ps, ucd )
	PARAMS ps ;
	unsigned char *ucd ;
{
	int x, x0, x1, i ;

	x0 = x1 = -1 ;
	for ( x=1; x<ps.xs-1; ++x ) {
		if ( ucd[x] == LOA && ucd[x-1] != LOA )
			x0 = x ;
		if ( ucd[x] == LOA && ucd[x+1] != LOA )
			x1 = x ;

		if ( x0>0 && x1>0 ) {
			for ( i=x0; i<=x1; ++i )
				if ( ucd[x0-1] > ucd[x1+1] )
					ucd[i] = ucd[x1+1] ;
				else
					ucd[i] = ucd[x0-1] ;
			x0 = x1 = -1 ;
		}

		if ( x0<0 && x1>0 ) {
			for ( i=0; i<=x1; ++i )
				ucd[i] = ucd[x1+1] ;
			x0 = x1 = -1 ;
		}

	}

	if ( x0>0 && x1<0 )
		for ( i=x0; i<ps.xs; ++i )
			ucd[i] = ucd[x0-1] ;

	return 0 ;
}

int write_img( ps, vs, t )
	PARAMS ps ;
	VARS vs ;
	int t ;
{
	char str[STR_LEN], *rf=ps.result_file ;
	int x, y, z, zmax, z0, z1 ;
	double max ;
	static unsigned char *ucd ;
	static FILE *fpu, *fpv, *fpr, *fpr2 ;

	if ( t==0 ) {
		ucd = (unsigned char *)malloc( ps.xs*sizeof*ucd ) ;

		fpr = fopen(strcat(strcpy(str,rf),".res"),"wb") ;

		if ( ps.sw_u==1 )
			fpu = fopen(strcat(strcpy(str,rf),".u"),"wb") ;

		if ( ps.sw_v==1 )
			fpv = fopen(strcat(strcpy(str,rf),".v"),"wb") ;

		if ( ps.sw_ip==1 )
			fpr2 = fopen(strcat(strcpy(str,rf),".res.ip"),"wb") ;


		if ( ucd==NULL || fpr==NULL
			|| ps.sw_u==1 && fpu==NULL
			|| ps.sw_v==1 && fpv==NULL
			|| ps.sw_ip==1 && fpr2==NULL ) {
			free( ucd ) ;
			fclose( fpr ) ;
			fclose( fpu ) ;
			fclose( fpv ) ;
			fclose( fpr2 ) ;
			return -1 ;
		}
	}

	if ( ps.sw_u == 1 )
	for ( z=0; z<ps.zs; ++z )
	for ( y=0; y<ps.ys; ++y ) {
		for ( x=0; x<ps.xs; ++x )
			ucd[x] = d2uc( vs.nu[z][y][x], U_LOW, U_HIGH ) ;
		fwrite( ucd, sizeof*ucd, ps.xs, fpu ) ;
	}

	if ( ps.sw_v == 1 )
	for ( z=0; z<ps.zs; ++z )
	for ( y=0; y<ps.ys; ++y ) {
		for ( x=0; x<ps.xs; ++x )
			ucd[x] = d2uc( vs.nv[z][y][x], V_LOW, V_HIGH ) ;
		fwrite( ucd, sizeof*ucd, ps.xs, fpv ) ;
	}

	for ( y=0; y<ps.ys; ++y ) {
		for ( x=0; x<ps.xs; ++x ) {
			zmax = 0 ;
			max = vs.nu[0][y][x] ;
			for ( z=1; z<ps.zs; ++z )
				if ( vs.nu[z][y][x] > max ) {
					zmax = z ;
					max = vs.nu[z][y][x] ;
				}

			ucd[x] = (unsigned char)zmax ;
		}

		fwrite( ucd, sizeof*ucd, ps.xs, fpr ) ;

		if ( ps.sw_ip == 1 ) {
			for ( x=0; x<ps.xs; ++x )
				if ( t>0 && vs.nu[ucd[x]][y][x] <= 0.5 )
					ucd[x] = LOA ;

			ip4loa( ps, ucd ) ;
			fwrite( ucd, sizeof*ucd, ps.xs, fpr2 ) ;
		}
	}

	if ( t == ps.ts ) {
		free( ucd ) ;
		fclose( fpr ) ;
		fclose( fpu ) ;
		fclose( fpv ) ;
		fclose( fpr2 ) ;
	}

	return 0 ;
}

int put_msg( ps )
	PARAMS ps ;
{
	printf( "PARAMS: \n parameter file; %s, result file; %s \n",
			ps.para_file, ps.result_file ) ;
	printf( " left=%s\n right=%s\n", ps.fnl, ps.fnr ) ;
	printf( " dx=%d, dy=%d, dz=%d, dt=%d\n", ps.dx, ps.dy, ps.dz, ps.dt ) ;
	printf( " lx=%d, ly=%d, lt=%d\n", ps.lx, ps.ly, ps.lt ) ;
	printf( " lz_min=%d, lz_max=%d\n", ps.lz_min, ps.lz_max ) ;
	printf( " xs=%d, ys=%d, zs=%d, ts=%d\n", ps.xs, ps.ys, ps.zs, ps.ts ) ;
	printf( " subpd=%d\n", ps.subpd ) ;
	printf( " Du=%e, Dv=%e\n a0=%f, a1=%f, a2=%f, b=%f, eps=%e\n",
			ps.Du, ps.Dv, ps.a0, ps.a1, ps.a2, ps.b, ps.eps ) ;
	printf( " mu_es=%f, mu_ini=%f\n", ps.mu_es, ps.mu_ini ) ;
	printf( " dt/dx^2 = %e, dt/dy^2=%e, dt/dz^2=%e\n",
			(double)(ps.dx*ps.dx)/(double)ps.dt,
			(double)(ps.dy*ps.dy)/(double)ps.dt,
			(double)(ps.dz*ps.dz)/(double)ps.dt ) ;
	printf( " Loop=%d, Delta=%e\n", ps.loop, ps.delta ) ;
	printf( " Interval: tintv=%d\n", ps.tintv ) ;
	printf( " Switch: sw_u=%d, sw_v=%d, sw_ip\n",
			ps.sw_u, ps.sw_v, ps.sw_ip ) ;
	printf( " nthr=%d\n", ps.nthr ) ;

	printf( " U_LOW=%f, U_HIGH=%f\n", U_LOW, U_HIGH ) ;
	printf( " V_LOW=%f, V_HIGH=%f\n", V_LOW, V_HIGH ) ;
	printf( " LOA=%d, tht=%f\n", LOA, ps.tht ) ;

	fflush( stdout ) ;

	return 0 ;
}

double chk_dpmap( ps, dpmap, z, y, x, hsubpd )
	PARAMS ps ;
	unsigned char **dpmap ;
	int z, y, x, hsubpd ;
{
	if ( x>0 )
	if ( dpmap[y][x-1] != LOA )
	if ( z>=(dpmap[y][x-1]-hsubpd) && z<=(dpmap[y][x-1]+hsubpd) )
		return 1. ;

	if ( x<ps.xs-1 )
	if ( dpmap[y][x+1] != LOA )
	if ( z>=(dpmap[y][x+1]-hsubpd) && z<=(dpmap[y][x+1]+hsubpd) )
		return 1. ;

	if ( y>0 )
	if ( dpmap[y-1][x] != LOA )
	if ( z>=(dpmap[y-1][x]-hsubpd) && z<=(dpmap[y-1][x]+hsubpd) )
		return 1. ;

	if ( y<ps.ys-1 )
	if ( dpmap[y+1][x] != LOA )
	if ( z>=(dpmap[y+1][x]-hsubpd) && z<=(dpmap[y+1][x]+hsubpd) )
		return 1. ;

	if ( y>0 && x>0 )
	if ( dpmap[y-1][x-1] != LOA )
	if ( z>=(dpmap[y-1][x-1]-hsubpd) && z<=(dpmap[y-1][x-1]+hsubpd) )
		return 1. ;

	if ( y>0 && x<ps.xs-1 )
	if ( dpmap[y-1][x+1] != LOA )
	if ( z>=(dpmap[y-1][x+1]-hsubpd) && z<=(dpmap[y-1][x+1]+hsubpd) )
		return 1. ;

	if ( y<ps.ys-1 && x>0 )
	if ( dpmap[y+1][x-1] != LOA )
	if ( z>=(dpmap[y+1][x-1]-hsubpd) && z<=(dpmap[y+1][x-1]+hsubpd) )
		return 1. ;

	if ( y<ps.ys-1 && x<ps.xs-1 )
	if ( dpmap[y+1][x+1] != LOA )
	if ( z>=(dpmap[y+1][x+1]-hsubpd) && z<=(dpmap[y+1][x+1]+hsubpd) )
		return 1. ;

	return 0. ;
}

void *mk_cuv( pthr )
	void *pthr ;
{
	PARAMS_PTH *ppth=(PARAMS_PTH *)pthr ;
	PARAMS ps = ppth->ps ;
	VARS vs = ppth->vs ;
	int yl = ppth->lo ;
	int yh = ppth->hi ;
	int hsubpd = ps.subpd/2 ;
	double ***ou = vs.ou, ***ov = vs.ov ;
	double *ou4zm0, *ou4zm1 ;

	int x, y, z, zl, zh ;
	double a, max, ratio ;

	ou4zm0 = (double *)malloc( ps.zs*sizeof(*ou4zm0) ) ;
	ou4zm1 = (double *)malloc( ps.zs*sizeof(*ou4zm1) ) ;

	while( 1 ) {
		sema_wait( &(ppth->start) ) ;

		for ( y=yl; y<yh; ++y )
		for ( x=0; x<ps.xs; ++x ) {
			for ( z=0; z<ps.zs; ++z ) {
				ou[z][y][x] = vs.nu[z][y][x] ;
				ov[z][y][x] = vs.nv[z][y][x] ;
			}

			ou4zm0[0] = ou[0][y][x] ;
			for ( z=1; z<ps.zs; ++z )
				if ( ou4zm0[z-1] < ou[z][y][x] )
					ou4zm0[z] = ou[z][y][x] ;
				else
					ou4zm0[z] = ou4zm0[z-1] ;

			ou4zm1[ps.zs-1] = ou[ps.zs-1][y][x] ;
			for ( z=ps.zs-2; z>=0; --z )
				if ( ou4zm1[z+1] < ou[z][y][x] )
					ou4zm1[z] = ou[z][y][x] ;
				else
					ou4zm1[z] = ou4zm1[z+1] ;

			for ( z=0; z<ps.zs; ++z ) {
				zl = z-hsubpd ;
				zh = z+hsubpd ;

				if ( zl <= 0 )
					max = ou4zm1[zh+1] ;
				else if ( zh >= ps.zs-1 )
					max = ou4zm0[zl-1] ;
				else if ( ou4zm0[zl-1] < ou4zm1[zh+1] )
					max = ou4zm1[zh+1] ;
				else
					max = ou4zm0[zl-1] ;

				a = ps.a0 +
					(ps.a2-ps.a0)*(1.+tanh(max-ps.a1))/2. ;
				vs.du[z][y][x] = ou[z][y][x]
					+ fuv(ou[z][y][x],ov[z][y][x],a)
						/ps.eps/ps.dt
					+ ps.mu_es*vs.es_sub[z][y][x]
						*vs.es[z][y][x]/ps.dt ;

				vs.dv[z][y][x] = ov[z][y][x]
					+ guv(ou[z][y][x],ov[z][y][x],ps.b)
						/ps.dt ;
			}
		}

		sema_post( &(ppth->stop) ) ;
	}
}

void *solve( pthr )
	void *pthr ;
{
	PARAMS_PTH *ppth = (PARAMS_PTH *)pthr ;
	PARAMS ps = ppth->ps ;
	VARS vs = ppth->vs ;
	int zl = ppth->lo, zl2 ;
	int zh = ppth->hi, zh2 ;

	double ***n, **nz0, **nzm, **nzp, ***d, **dz0 ;
	double p, Cx, Cy, Cz ;

	int x, y, z, xs=ps.xs, ys=ps.ys, zs=ps.zs ;
	double sum, td ;

	if ( ppth->fl == DO_SOLVE_U ) {
		n = vs.nu ;
		d = vs.du ;
		p = ps.pu ;
		Cx = ps.Cux ;
		Cy = ps.Cuy ;
		Cz = ps.Cuz ;
	} else if ( ppth->fl == DO_SOLVE_V ) {
		n = vs.nv ;
		d = vs.dv ;
		p = ps.pv ;
		Cx = ps.Cvx ;
		Cy = ps.Cvy ;
		Cz = ps.Cvz ;
	} else {
		fprintf( stderr, "error: fl=%d\n", ppth->fl ) ;
		exit( 3 ) ;
	}

	if ( zl<=0 ) zl2 = 1 ; else zl2 = zl ;
	if ( zh>=zs-1 ) zh2 = zs-2 ; else zh2 = zh ;

	while ( 1 ) {
		sema_wait( &(ppth->start) ) ;

		sum = .0 ;

		/******** z=0 *******************************/
		if ( zl==0 ) {
			nz0 = n[0] ; nzp = n[1] ;
			dz0 = d[0] ;

			td =	( Cx*nz0[0][1]
				+ Cy*nz0[1][0]
				+ Cz*nzp[0][0]
				+ dz0[0][0] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[0][0]-td ) ;
			nz0[0][0] = td ;

			for ( x=1; x<xs-1; ++x ) {
				td =	( Cx*(nz0[0][x+1]+nz0[0][x-1])
					+ Cy*nz0[1][x]
					+ Cz*nzp[0][x]
					+ dz0[0][x] ) / (p-Cy-Cz) ;
				sum += fabs( nz0[0][x]-td ) ;
				nz0[0][x] = td ;
			}

			td =	( Cx*nz0[0][xs-2]
				+ Cy*nz0[1][xs-1]
				+ Cz*nzp[0][xs-1]
				+ dz0[0][xs-1] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[0][xs-1]-td ) ;
			nz0[0][xs-1] = td ;

			for ( y=1; y<ys-1; ++y ) {
				td =	( Cx*nz0[y][1]
					+ Cy*(nz0[y+1][0]+nz0[y-1][0])
					+ Cz*nzp[y][0]
					+ dz0[y][0] ) / (p-Cx-Cz) ;
				sum += fabs( nz0[y][0]-td ) ;
				nz0[y][0] = td ;

				for ( x=1; x<xs-1; ++x ) {
					td =	( Cx*(nz0[y][x+1]+nz0[y][x-1])
						+ Cy*(nz0[y+1][x]+nz0[y-1][x])
						+ Cz*nzp[y][x]
						+ dz0[y][x] ) / (p-Cz) ;
					sum += fabs( nz0[y][x]-td ) ;
					nz0[y][x] = td ;
				}

				td =	( Cx*nz0[y][xs-2]
					+ Cy*(nz0[y+1][xs-1]+nz0[y-1][xs-1])
					+ Cz*nzp[y][xs-1]
					+ dz0[y][xs-1] ) / (p-Cx-Cz) ;
				sum += fabs( nz0[y][xs-1]-td ) ;
				nz0[y][xs-1] = td ;
			}

			td =	( Cx*nz0[ys-1][1]
				+ Cy*nz0[ys-2][0]
				+ Cz*nzp[ys-1][0]
				+ dz0[ys-1][0] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[ys-1][0]-td ) ;
			nz0[ys-1][0] = td ;

			for ( x=1; x<xs-1; ++x ) {
				td =	( Cx*(nz0[ys-1][x+1]+nz0[ys-1][x-1])
					+ Cy*nz0[ys-2][x]
					+ Cz*nzp[ys-1][x]
					+ dz0[ys-1][x] ) / (p-Cy-Cz) ;
				sum += fabs( nz0[ys-1][x]-td ) ;
				nz0[ys-1][x] = td ;
			}

			td =	( Cx*nz0[ys-1][xs-2]
				+ Cy*nz0[ys-2][xs-1]
				+ Cz*nzp[ys-1][xs-1]
				+ dz0[ys-1][xs-1] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[ys-1][xs-1]-td ) ;
			nz0[ys-1][xs-1] = td ;
		}

		/************** 0<z<zs-1 *******************/

		for ( z=zl2; z<zh2; ++z ) {
			nz0 = n[z] ; nzm = n[z-1]; nzp = n[z+1] ;
			dz0 = d[z] ;

			td =	( Cx*nz0[0][1]
				+ Cy*nz0[1][0]
				+ Cz*(nzp[0][0]+nzm[0][0])
				+ dz0[0][0] ) / (p-Cx-Cy) ;
			sum += fabs( nz0[0][0]-td ) ;
			nz0[0][0] = td ;

			for ( x=1; x<xs-1; ++x ) {
				td =	( Cx*(nz0[0][x+1]+nz0[0][x-1])
					+ Cy*nz0[1][x]
					+ Cz*(nzp[0][x]+nzm[0][x])
					+ dz0[0][x] ) / (p-Cy) ;
				sum += fabs( nz0[0][x]-td ) ;
				nz0[0][x] = td ;
			}

			td =	( Cx*nz0[0][xs-2]
				+ Cy*nz0[1][xs-1]
				+ Cz*(nzp[0][xs-1]+nzm[0][xs-1])
				+ dz0[0][xs-1] ) / (p-Cx-Cy) ;
			sum += fabs( nz0[0][xs-1]-td ) ;
			nz0[0][xs-1] = td ;

			for ( y=1; y<ys-1; ++y ) {
				td =	( Cx*nz0[y][1]
					+ Cy*(nz0[y+1][0]+nz0[y-1][0])
					+ Cz*(nzp[y][0]+nzm[y][0])
					+ dz0[y][0] ) / (p-Cx) ;
				sum += fabs( nz0[y][0]-td ) ;
				nz0[y][0] = td ;

				for ( x=1; x<xs-1; ++x ) {
					td =	( Cx*(nz0[y][x+1]+nz0[y][x-1])
						+ Cy*(nz0[y+1][x]+nz0[y-1][x])
						+ Cz*(nzp[y][x]+nzm[y][x])
						+ dz0[y][x] ) / p ;
					sum += fabs( nz0[y][x]-td ) ;
					nz0[y][x] = td ;
				}

				td =	( Cx*nz0[y][xs-2]
					+ Cy*(nz0[y+1][xs-1]+nz0[y-1][xs-1])
					+ Cz*(nzp[y][xs-1]+nzm[y][xs-1])
					+ dz0[y][xs-1] ) / (p-Cx) ;
				sum += fabs( nz0[y][xs-1]-td ) ;
				nz0[y][xs-1] = td ;
			}

			td =	( Cx*nz0[ys-1][1]
				+ Cy*nz0[ys-2][0]
				+ Cz*(nzp[ys-1][0]+nzm[ys-1][0])
				+ dz0[ys-1][0] ) / (p-Cx-Cy) ;
			sum += fabs( nz0[ys-1][0]-td ) ;
			nz0[ys-1][0] = td ;

			for ( x=1; x<xs-1; ++x ) {
				td =	( Cx*(nz0[ys-1][x+1]+nz0[ys-1][x-1])
					+ Cy*nz0[ys-2][x]
					+ Cz*(nzp[ys-1][x]+nzm[ys-1][x])
					+ dz0[ys-1][x] ) / (p-Cy) ;
				sum += fabs( nz0[ys-1][x]-td ) ;
				nz0[ys-1][x] = td ;
			}

			td =	( Cx*nz0[ys-1][xs-2]
				+ Cy*nz0[ys-2][xs-1]
				+ Cz*(nzp[ys-1][xs-1]+nzm[ys-1][xs-1])
				+ dz0[ys-1][xs-1] ) / (p-Cx-Cy) ;
			sum += fabs( nz0[ys-1][xs-1]-td ) ;
			nz0[ys-1][xs-1] = td ;
		}

		/********** z=zs-1 *************/

		if ( zh==zs-1 ) {
			nz0 = n[zs-1] ; nzm = n[zs-2] ;
			dz0 = d[zs-1] ;

			td =	( Cx*nz0[0][1]
				+ Cy*nz0[1][0]
				+ Cz*nzm[0][0]
				+ dz0[0][0] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[0][0]-td ) ;
			nz0[0][0] = td ;

			for ( x=1; x<xs-1; ++x ) {
				td =	( Cx*(nz0[0][x+1]+nz0[0][x-1])
					+ Cy*nz0[1][x]
					+ Cz*nzm[0][x]
					+ dz0[0][x] ) / (p-Cy-Cz) ;
				sum += fabs( nz0[0][x]-td ) ;
				nz0[0][x] = td ;
			}

			td =	( Cx*nz0[0][xs-2]
				+ Cy*nz0[1][xs-1]
				+ Cz*nzm[0][xs-1]
				+ dz0[0][xs-1] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[0][xs-1]-td ) ;
			nz0[0][xs-1] = td ;

			for ( y=1; y<ys-1; ++y ) {
				td =	( Cx*nz0[y][1]
					+ Cy*(nz0[y+1][0]+nz0[y-1][0])
					+ Cz*nzm[y][0]
					+ dz0[y][0] ) / (p-Cx-Cz) ;
				sum += fabs( nz0[y][0]-td ) ;
				nz0[y][0] = td ;

				for ( x=1; x<xs-1; ++x ) {
					td =	( Cx*(nz0[y][x+1]+nz0[y][x-1])
						+ Cy*(nz0[y+1][x]+nz0[y-1][x])
						+ Cz*nzm[y][x]
						+ dz0[y][x] ) / (p-Cz) ;
					sum += fabs( nz0[y][x]-td ) ;
					nz0[y][x] = td ;
				}

				td =	( Cx*nz0[y][xs-2]
					+ Cy*(nz0[y+1][xs-1]+nz0[y-1][xs-1])
					+ Cz*nzm[y][xs-1]
					+ dz0[y][xs-1] ) / (p-Cx-Cz) ;
				sum += fabs( nz0[y][xs-1]-td ) ;
				nz0[y][xs-1] = td ;
			}

			td =	( Cx*nz0[ys-1][1]
				+ Cy*nz0[ys-2][0]
				+ Cz*nzm[ys-1][0]
				+ dz0[ys-1][0] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[ys-1][0]-td ) ;
			nz0[ys-1][0] = td ;

			for ( x=1; x<xs-1; ++x ) {
				td =	( Cx*(nz0[ys-1][x+1]+nz0[ys-1][x-1])
					+ Cy*nz0[ys-2][x]
					+ Cz*nzm[ys-1][x]
					+ dz0[ys-1][x] ) / (p-Cy-Cz) ;
				sum += fabs( nz0[ys-1][x]-td ) ;
				nz0[ys-1][x] = td ;
			}

			td =	( Cx*nz0[ys-1][xs-2]
				+ Cy*nz0[ys-2][xs-1]
				+ Cz*nzm[ys-1][xs-1]
				+ dz0[ys-1][xs-1] ) / (p-Cx-Cy-Cz) ;
			sum += fabs( nz0[ys-1][xs-1]-td ) ;
			nz0[ys-1][xs-1] = td ;
		}

		ppth->sum_dif = sum ;

		sema_post( &(ppth->stop) ) ;
	}
}

int read_stereo_images( ps, vs )
	PARAMS ps ;
	VARS vs ;
{
	FILE *fp ;
	int y ;

	if ( (fp=fopen(ps.fnl,"rb")) == NULL ) {
		fprintf( stderr, "Error: cannot open file %s.\n", ps.fnl ) ;
		return -1 ;
	} else {
		for ( y=0; y<ps.ys; ++y )
			fread( vs.image_left[y], sizeof **vs.image_left,
				ps.xs, fp ) ;

		fclose( fp ) ;
	}

	if ( (fp=fopen(ps.fnr,"rb")) == NULL ) {
		fprintf( stderr, "Error: cannot open file %s.\n", ps.fnr ) ;
		return -1 ;
	} else {
		for ( y=0; y<ps.ys; ++y )
			fread( vs.image_right[y], sizeof **vs.image_right,
				ps.xs, fp ) ;

		fclose( fp ) ;
	}

	return 0 ;
}

double get_ipi( ps, x_dbl, y_int, image )
	PARAMS ps ;
	double x_dbl ;
	int y_int ;
	unsigned char **image ;
{
	if ( x_dbl<0. || y_int<0 || x_dbl>ps.xs-1 || y_int>ps.ys-1 )
		return -1. ;
	else {
		int x0 ;
		double xr, d ;

		x0 = (int)x_dbl ;
		xr = x_dbl-x0 ;

		d = (1.-xr)*image[y_int][x0] + xr*image[y_int][x0+1] ;

		return d ;
	}
}

double get_cor( ps, il, ir, xl, xr, y0, hbs )
	PARAMS ps ;
	unsigned char **il, **ir ;
	int xl, y0, hbs ;
	double xr ;
{
	int	n, hbs2=2*hbs+1 ;
	int	i, i_start, i_end ;
	double	j, j_start, j_end ;
	int	k, k_start, k_end ;
	double	avl, avr, sl, sr, s ;
	static	double *ipi ;

	if ( hbs>5 ) {
/*
		printf( "get_cor: xl=%d, xr=%f, y0=%d\n", xl, xr, y0 ) ;
*/
		return 0. ;
	}

	if ( ipi==NULL )
		ipi=(double *)malloc(hbs2*hbs2*sizeof*ipi) ;

	if ( ( i_start=xl-hbs ) < 0 ) i_start = 0 ;
	if ( ( i_end=xl+hbs ) > ps.xs-1 ) i_end = ps.xs-1 ;

	if ( ( j_start=xr-hbs ) < 0 ) j_start = 0 ;
	if ( ( j_end=xr+hbs ) > ps.xs-1 ) j_end = ps.xs-1 ;

	if ( ( k_start=y0-hbs ) < 0 ) k_start = 0 ;
	if ( ( k_end=y0+hbs ) > ps.ys-1 ) k_end = ps.ys-1 ;

	avl = avr = .0 ;
	n = 0 ;
	for ( k=k_start; k<=k_end; ++k )
	for ( i=i_start, j=j_start; i<=i_end && j<=j_end; ++i, j+=1. ) {
		avl += il[k][i] ;
		ipi[n] = get_ipi( ps, j, k, ir ) ;
		avr += ipi[n++] ;
	}

	avl /= n ;
	avr /= n ;

	sl = sr = s = .0 ;
	n = 0 ;
	for ( k=k_start; k<=k_end; ++k )
	for ( i=i_start, j=j_start; i<=i_end && j<=j_end; ++i, j+=1. ) {
		s  += (il[k][i]-avl)*(ipi[n]-avr) ;
		sl += pow( il[k][i]-avl, 2. ) ;
		sr += pow( ipi[n++]-avr, 2. ) ;
	}

	if ( s>0. && ( sl > 0. || sr > 0. ) )
		return s / ( sl>sr ? sl: sr ) ;
	else
		return get_cor( ps, il, ir, xl, xr, y0, hbs+1 ) ;
}

int set_stimuli( ps, vs )
	PARAMS ps ;
	VARS vs ;
{
	int x, y, z, xl ;
	double xr, avl, avr, s, sl, sr, cor ;
	double ir0, ir1, ir2, ir3, ir4 ;
	unsigned char **il=vs.image_left, **ir=vs.image_right ;

	for ( z=0; z<ps.zs; ++z ) {
		for ( y=0; y<ps.ys; ++y )
		for ( x=0; x<ps.xs; ++x )
			vs.es[z][y][x] = 0. ;

		for ( y=1; y<ps.ys-1; ++y )
		for ( x=1; x<ps.xs-1; ++x ) {
			xl = x ;
			xr = xl-(double)z/ps.subpd-ps.lz_min ;
			if ( xr<1. ) continue ;

			ir0 = get_ipi(ps,xr,   y,  ir) ;
			ir1 = get_ipi(ps,xr-1.,y,  ir) ;
			ir2 = get_ipi(ps,xr+1.,y,  ir) ;
			ir3 = get_ipi(ps,xr,   y-1,ir) ;
			ir4 = get_ipi(ps,xr,   y+1,ir) ;

			sr  = ir0 ;
			sr += ir1 ;
			sr += ir2 ;
			sr += ir3 ;
			sr += ir4 ;
			avr = sr/5. ;

			sl  = il[y][xl] ;
			sl += il[y][xl-1] ;
			sl += il[y][xl+1] ;
			sl += il[y-1][xl] ;
			sl += il[y+1][xl] ;
			avl = sl/5. ;

			s   = (ir0-avr)*(il[y][xl]  -avl) ;
			s  += (ir1-avr)*(il[y][xl-1]-avl) ;
			s  += (ir2-avr)*(il[y][xl+1]-avl) ;
			s  += (ir3-avr)*(il[y-1][xl]-avl) ;
			s  += (ir4-avr)*(il[y+1][xl]-avl) ;

			sr  = pow( ir0-avr, 2.) ;
			sr += pow( ir1-avr, 2.) ;
			sr += pow( ir2-avr, 2.) ;
			sr += pow( ir3-avr, 2.) ;
			sr += pow( ir4-avr, 2.) ;

			sl  = pow( il[y][xl]  -avl, 2.) ;
			sl += pow( il[y][xl-1]-avl, 2.) ;
			sl += pow( il[y][xl+1]-avl, 2.) ;
			sl += pow( il[y-1][xl]-avl, 2.) ;
			sl += pow( il[y+1][xl]-avl, 2.) ;

			if ( s>.0 && ( sr > 0.0 || sl > 0.0 ) )
				vs.es[z][y][x] = s / ( sr>sl ? sr: sl ) ;
			else
				vs.es[z][y][x] = .0 ;
		}
	}

	return 0 ;
}

int set_initial_conditions( ps, vs )
	PARAMS ps ;
	VARS vs ;
{
	int x, y, z ;

	for ( z=0; z<ps.zs; ++z )
	for ( y=0; y<ps.ys; ++y )
	for ( x=0; x<ps.xs; ++x ) {
		vs.ou[z][y][x] = vs.nu[z][y][x] = ps.mu_ini*vs.es[z][y][x] ;
		vs.ov[z][y][x] = vs.nv[z][y][x] = .0 ;
		vs.es_sub[z][y][x] = 1.0 ;
	}

	return 0 ;
}

int estimate_esmu( ps, vs )
	PARAMS ps ;
	VARS vs ;
{
	int x, y, z, zmax ;
	double max ;

	for ( y=0; y<ps.ys; ++y )
	for ( x=0; x<ps.xs; ++x ) {
		zmax = 0 ;
		max = vs.nu[0][y][x] ;
		for ( z=1; z<ps.zs; ++z )
			if ( vs.nu[z][y][x] > max ) {
				zmax = z ;
				max = vs.nu[z][y][x] ;
			}

		if ( max <= 0.5 )
			vs.dpmap[y][x] = LOA ;
		else
			vs.dpmap[y][x] = (unsigned char)zmax ;
	}

	for ( y=0; y<ps.ys; ++y )
	for ( x=ps.lz_max+1; x<ps.xs; ++x )
		if ( vs.dpmap[y][x] == LOA )
		for ( z=0; z<ps.zs; ++z )
			if ( vs.es_sub[z][y][x]==1.0 )
				vs.es_sub[z][y][x] +=
				chk_dpmap(ps,vs.dpmap,z,y,x,ps.subpd/2) ;

	return 0 ;
}

int estimate_esmu2( ps, vs )
	PARAMS ps ;
	VARS vs ;
{
	int x, y, xl, xr ;
	int z, zmin, zmax ;

	for ( y=0; y<ps.ys; ++y )
	for ( x=ps.lz_max+1; x<ps.xs; ++x )
		if ( vs.dpmap[y][x] == LOA ) {
			for ( xl=x; vs.dpmap[y][xl]==LOA && xl>=0; --xl ) ;
			for ( xr=x; vs.dpmap[y][xr]==LOA && xr<ps.xs; ++xr ) ;

			if ( xl<0 ) xl=xr ;
			if ( xr>ps.xs-1 ) xr=xl ;

			if ( vs.dpmap[y][xl] > vs.dpmap[y][xr] ) {
				zmin = vs.dpmap[y][xr] ;
				zmax = vs.dpmap[y][xl] ;
			} else {
				zmin = vs.dpmap[y][xl] ;
				zmax = vs.dpmap[y][xr] ;
			}
			
			for ( z=zmin; z<=zmax; ++z )
				if ( vs.es_sub[z][y][x] == 1.0 )
					vs.es_sub[z][y][x] += 1.0 ;
		}

	return 0 ;
}

int calc( ps, vs )
	PARAMS ps ;
	VARS vs ;
{
	int x, y, z, t, n, l, yl, yh, zl, zh ;
	double max, sum_dif, delta_xyzs = ps.delta*ps.xyzs ;

	PARAMS_PTH *thr_mk_cuv=(PARAMS_PTH *)
			malloc(ps.nthr*sizeof*thr_mk_cuv) ;
	PARAMS_PTH *thr_solve_u=(PARAMS_PTH *)
			malloc(ps.nthr*sizeof*thr_solve_u) ;
	PARAMS_PTH *thr_solve_v=(PARAMS_PTH *)
			malloc(ps.nthr*sizeof*thr_solve_v) ;

	if ( put_msg( ps ) == -1 )
		return -1 ;
	else if ( read_stereo_images( ps, vs ) == -1 )
		return -1 ;
	else if ( set_stimuli( ps, vs ) == -1 )
		return -1 ;
	else if ( set_initial_conditions( ps, vs ) == -1 )
		return -1 ;
	else
		write_img( ps, vs, 0 ) ;

	for ( n=0; n<ps.nthr; ++n ) {
		yl = n*ps.ys/ps.nthr ;
		yh = (n+1)*ps.ys/ps.nthr ;

		zl = n*ps.zs/ps.nthr ;
		zh = (n+1)*ps.zs/ps.nthr ;

		thr_mk_cuv[n].ps = ps ;
		thr_mk_cuv[n].vs = vs ;
		thr_mk_cuv[n].lo = yl ;
		thr_mk_cuv[n].hi = yh ;
		thr_mk_cuv[n].fl = DO_MK_CUV ;
		sema_init( &(thr_mk_cuv[n].start), 0, USYNC_THREAD, NULL ) ;
		sema_init( &(thr_mk_cuv[n].stop) , 0, USYNC_THREAD, NULL ) ;
		thr_create( NULL, NULL, mk_cuv, &(thr_mk_cuv[n]),
				THR_BOUND, &(thr_mk_cuv[n].tid) ) ;

		thr_solve_u[n].ps = ps ;
		thr_solve_u[n].vs = vs ;
		thr_solve_u[n].lo = zl ;
		thr_solve_u[n].hi = zh ;
		thr_solve_u[n].fl = DO_SOLVE_U ;
		sema_init( &(thr_solve_u[n].start), 0, USYNC_THREAD, NULL ) ;
		sema_init( &(thr_solve_u[n].stop) , 0, USYNC_THREAD, NULL ) ;
		thr_create( NULL, NULL, solve, &(thr_solve_u[n]),
				THR_BOUND, &(thr_solve_u[n].tid) ) ;

		thr_solve_v[n].ps = ps ;
		thr_solve_v[n].vs = vs ;
		thr_solve_v[n].lo = zl ;
		thr_solve_v[n].hi = zh ;
		thr_solve_v[n].fl = DO_SOLVE_V ;
		sema_init( &(thr_solve_v[n].start), 0, USYNC_THREAD, NULL ) ;
		sema_init( &(thr_solve_v[n].stop) , 0, USYNC_THREAD, NULL ) ;
		thr_create( NULL, NULL, solve, &(thr_solve_v[n]),
				THR_BOUND, &(thr_solve_v[n].tid) ) ;
	}

	for ( t=1; t<ps.ts; ++t ) {
		if ( t%ps.tintv==0 ) {
			printf( "----- t=%f\n", (double)t/(double)ps.dt ) ;
			printf( "      %e, %e\n",
				vs.ou[0][ps.ys/2][ps.xs/2],
				vs.ov[0][ps.ys/2][ps.xs/2] ) ;
			fflush( stdout ) ;
			write_img( ps, vs, t ) ;
		}

		for ( n=0; n<ps.nthr; ++n )
			sema_post( &(thr_mk_cuv[n].start) ) ;
		for ( n=0; n<ps.nthr; ++n )
			sema_wait( &(thr_mk_cuv[n].stop)  ) ;

		l = ps.loop ;
		while( l-- > 0 ) {
			for ( n=0; n<ps.nthr; ++n ) {
				sema_post( &(thr_solve_u[n].start) ) ;
				sema_post( &(thr_solve_v[n].start) ) ;
			}
			sum_dif = .0 ;
			for ( n=0; n<ps.nthr; ++n ) {
				sema_wait( &(thr_solve_u[n].stop) ) ;
				sema_wait( &(thr_solve_v[n].stop) ) ;
				sum_dif += thr_solve_u[n].sum_dif ;
			}
			if ( sum_dif<delta_xyzs )
				break ;
		}

	 	if ( l<=0 ) puts( "exceeded the max loop number" ) ;

		if ( t >= (int)(ps.tht*ps.dt) ) estimate_esmu( ps, vs ) ;

		if ( t == (int)(ps.tht*ps.dt) ) estimate_esmu2( ps, vs ) ;
	}

	write_img( ps, vs, ps.ts ) ;

	return 0 ;
}

