/*
	eval_disp6a.c (previous: eval_disp6.c)
	evaluate a disparity map
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <an.h>

#define STR_LEN	512

typedef struct {
	double fac, loa, dlt, dz ;
	unsigned char *tdp ;	/* groundtruth: groundtruth disparity map */
	unsigned char *noc ;	/* nonocc: non-occlusion area */
	unsigned char *all ;	/* all: all area */
	unsigned char *ddi ;	/* disc: depth discontinuity area:
					  0: occlusion
					128: normal area
					255: depth discontinuity area */
	unsigned char *odp ;	/* obtained disparity map */
	FILE *fp_odp ;
} PARAMS ;


FILE *df_open( fp, mode )
	FILE *fp ;
	char *mode ;
{
	char str[STR_LEN], *p ;

	fgets( str, STR_LEN, fp ) ;

	for ( p=str; *p!='\n' && *p!='\r' && *p!=';'
			&& *p!=' ' && *p!='\t' && *p!=NULL ; ++p ) ;
	*p = (char)NULL ;

	printf( "# file=%s\n", str ) ;
	return fopen( str, mode ) ;
}

int get_params( fn, pps, img )
	char *fn ;
	PARAMS *pps ;
	AN_IMAGE_COORD img ;
{
	char str[STR_LEN] ;
	FILE *fp, *fp2 ;
	int s = img.x * img.y ;

	if ( ( fp = fopen( fn, "r" ) ) == NULL ) {
		fputs( "error: cannot open a file.\n", stderr ) ;
		return -1 ;
	}

	pps->fac = atof( fgets( str, STR_LEN, fp ) ) ;
	pps->loa = atof( fgets( str, STR_LEN, fp ) ) ;
	pps->dlt = atof( fgets( str, STR_LEN, fp ) ) ;

	if ( ( fp2 = df_open( fp, "rb" ) ) == NULL )	/* true disparity */
		return -1 ;
	else {
		fread( pps->tdp, sizeof *(pps->tdp), s, fp2 ) ;
		fclose( fp2 ) ;
	}

	if ( ( fp2 = df_open( fp, "rb" ) ) == NULL )	/* non-occlusion */
		return -1 ;
	else {
		fread( pps->noc, sizeof *(pps->noc), s, fp2 ) ;
		fclose( fp2 ) ;
	}

	if ( ( fp2 = df_open( fp, "rb" ) ) == NULL )	/* all */
		return -1 ;
	else {
		fread( pps->all, sizeof *(pps->all), s, fp2 ) ;
		fclose( fp2 ) ;
	}

	if ( ( fp2 = df_open( fp, "rb" ) ) == NULL )	/* discontinuity */
		return -1 ;
	else {
		fread( pps->ddi, sizeof *(pps->ddi), s, fp2 ) ;
		fclose( fp2 ) ;
	}

	if ( ( pps->fp_odp = df_open( fp, "rb" ) ) == NULL )
		return -1 ;
						/* obtained disparity map */

	pps->dz = atof( fgets( str, STR_LEN, fp ) ) ;

	fclose( fp ) ;

	return 0 ;
}

int get_dif( ps, dif, img )
	PARAMS ps ;
	double *dif ;
	AN_IMAGE_COORD img ;
{
	int p ;

	for ( p=0; p<img.x*img.y; ++p )
		if ( ps.odp[p]!=ps.loa && ps.all[p]==255 )
			dif[p] = fabs( (double)ps.tdp[p]/ps.fac
					- (double)ps.odp[p]/ps.dz ) ;
		else
			dif[p] = .0 ;

	return 0 ;
}

int eval_nocc( ps, dif, img )
	PARAMS ps ;
	double *dif ;
	AN_IMAGE_COORD img ;
{
	int c, f, p ;
	double sum ;

	c=f=0 ;
	sum=.0 ;
	for ( p=0; p<img.x*img.y; ++p )
		if ( ps.odp[p]!=ps.loa && ps.noc[p]==255 ) {
			++c ;
			if ( dif[p]>ps.dlt ) ++f ;
			sum += pow( dif[p], 2. ) ;
		}

	if ( c>0 )
		printf( "BMP-noc=%e, RMS-noc=%e (c=%d)\n",
			(double)f/(double)c*100.,
			sqrt( sum/(double)c ), c ) ;
	else
		printf( "BMP-noc=-, RMS-noc=- (c=0)\n" ) ;

	return 0 ;
}

int eval_ddi( ps, dif, img )
	PARAMS ps ;
	double *dif ;
	AN_IMAGE_COORD img ;
{
	int c, f, p ;
	double sum ;

	c = f = 0 ;
	sum = .0 ;
	for ( p=0; p<img.x*img.y; ++p )
		if ( ps.odp[p]!=ps.loa && ps.ddi[p]==255 ) {
			++c ;
			if ( dif[p]>ps.dlt ) ++f ;
			sum += pow( dif[p], 2. ) ;
		}

	if ( c>0 )
		printf( "BMP-ddi=%e, RMS-ddi=%e (c=%d)\n",
			(double)f/(double)c*100.,
			sqrt( sum/(double)c ), c ) ;
	else
		printf( "BMP-ddi=-, RMS-ddi=- (c=0)\n" ) ;

	return 0 ;
}

int eval_all( ps, dif, img )
	PARAMS ps ;
	double *dif ;
	AN_IMAGE_COORD img ;
{
	int c, f, p ;
	double sum ;

	c = f = 0 ;
	sum = .0 ;
	for ( p=0; p<img.x*img.y; ++p )
		if ( ps.odp[p]!=ps.loa && ps.all[p]==255 ) {
			++c ;
			if ( dif[p]>ps.dlt ) ++f ;
			sum += pow( dif[p], 2. ) ;
		}

	if ( c>0 )
		printf( "BMP-all=%e, RMS-all=%e (c=%d)\n",
			(double)f/(double)c*100.,
			sqrt( sum/(double)c ), c ) ;
	else
		printf( "BMP-all=-, RMS-all=- (c=0)\n" ) ;

	return 0 ;
}

int eval_disp( ps, fn, img )
	PARAMS ps ;
	char *fn ;
	AN_IMAGE_COORD img ;
{
	int fs, t ;
	double *dif ;
	char str[STR_LEN] ;
	FILE *fp ;

	fs = img.x*img.y ;
	fp = fopen( strcat( strcpy( str, fn ), ".dif" ), "wb" ) ;
	dif = (double *)malloc( sizeof *dif * fs ) ;

	printf( "# size: x=%d, y=%d, t=%d\n", img.x, img.y, img.t ) ;
	printf( "# fac=%f, loa=%f, dlt=%f\n", ps.fac, ps.loa, ps.dlt ) ;
	printf( "# dz=%f\n", ps.dz ) ;

	for ( t=0; t<img.t; ++t ) {
		fread( ps.odp, sizeof *(ps.odp), fs, ps.fp_odp ) ;

		printf( "----- t=%d -----\n", t ) ;
		get_dif( ps, dif, img ) ;
		eval_nocc( ps, dif, img ) ;
		eval_all( ps, dif, img ) ;
		eval_ddi( ps, dif, img ) ;

		fwrite( dif, sizeof *dif, fs, fp ) ;
	}

	fclose( fp ) ;

	return 0 ;
}

int main( argc, argv )
	int argc ;
	char **argv ;
{
	AN_IMAGE_COORD img ;
	PARAMS ps ;

	if ( argc != 3 ) {
		fprintf( stderr, "Usage: %s data_file dif_file\n",
			argv[0] ) ;
		return( 1 ) ;
	} else if ( an_get_img( &img ) == -1 ) {
		fputs( "Error: cannot get image environment.\n", stderr ) ;
		return( 2 ) ;
	} else if ( (ps.tdp=(unsigned char *)
			malloc(img.x*img.y*sizeof(*(ps.tdp))))==NULL ||
		(ps.noc=(unsigned char *)
			malloc(img.x*img.y*sizeof(*(ps.noc))))==NULL ||
		(ps.all=(unsigned char *)
			malloc(img.x*img.y*sizeof(*(ps.all))))==NULL ||
		(ps.ddi=(unsigned char *)
			malloc(img.x*img.y*sizeof(*(ps.ddi))))==NULL ||
		(ps.odp=(unsigned char *)
			malloc(img.x*img.y*sizeof(*(ps.odp))))==NULL ) {
		fputs( "Error: cannot allocate memory.\n", stderr ) ;
		free( ps.tdp ) ; free( ps.noc ) ;
		free( ps.all ) ; free( ps.ddi ) ; free( ps.odp ) ;
		return( 2 ) ;
	} else if ( get_params( argv[1], &ps, img ) == -1 ) {
		fputs( "error in get_params()\n", stderr ) ;
		free( ps.tdp ) ; free( ps.noc ) ;
		free( ps.all ) ; free( ps.ddi ) ; free( ps.odp ) ;
		return( 2 ) ;
	} else if ( eval_disp( ps, argv[2], img ) == -1 ) {
		fputs( "Error in eval_disp()\n", stderr ) ;
		free( ps.tdp ) ; free( ps.noc ) ;
		free( ps.all ) ; free( ps.ddi ) ; free( ps.odp ) ;
		fclose( ps.fp_odp ) ;
		return( 2 ) ;
	} else {
		free( ps.tdp ) ; free( ps.noc ) ;
		free( ps.all ) ; free( ps.ddi ) ; free( ps.odp ) ;
		fclose( ps.fp_odp ) ;
		return( 0 ) ;
	}
}

