#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <an.h>

double inpro( a, b, n )
	double *a, *b ;
	int n ;
{
	double s ;

	for ( --n, s=.0; n>=0; --n )
		s += a[n]*b[n] ;

	return s ;
}

static int swap( matrix, n, size )
	double **matrix ;
	int n, size ;
{
	int i, j, mi ;
	double max, tmp ;

	for ( max=.0, mi=n, i=n+1 ; i<size ; ++i )
		if ( (tmp=fabs(matrix[i][n]))>max )
			max=tmp, mi=i ;

	if ( mi==n )
		return -1 ;

	for ( i=n ; i<size+1 ; ++i ) {
		double tmp = matrix[mi][i] ;
		matrix[mi][i] = matrix[n][i] ;
		matrix[n][i] = tmp ;
	}

	return 0 ;
}

static int scaling( matrix, size )
	double **matrix ;
	int size ;
{
	int i, j ;
	double tmp, tmp2 ;

	for ( i=0 ; i<size ; ++i ) {
		for ( tmp=.0, j=0 ; j<size ; ++j )
			if ( tmp<(tmp2=fabs(matrix[i][j])) )
				tmp = tmp2 ;
		if ( tmp==.0 )
			return -1 ;
		else
			for ( j=0 ; j<size+1 ; ++j )
				matrix[i][j] /= tmp ;
	}

	return 0 ;
}

int gauss_jordan( matrix, size )
	double **matrix ;
	int size ;
{
	int i , j , k ;
	double tmpm ;

	if ( scaling( matrix, size ) == -1 )
		return -1 ;

	for ( j = 0 ; j < size ; ++j ) {
		if ( matrix[j][j] == .0 &&
			swap( matrix , j, size ) == -1 )
			return -1 ;
		for ( tmpm=matrix[j][j] , i = 0 ; i < size + 1 ; ++i )
			matrix[j][i] /= tmpm ;
		for ( i = 0 ; i < size ; ++i ) {
			double tmp = matrix[i][j] ;
			if ( i != j )
				for ( k = 0 ; k < size + 1 ; ++k )
					matrix[i][k] -=
						matrix[j][k] * tmp ;
		}
	}

	return 0 ;
}

int gauss( matrix, size )
	double **matrix ;
	int size ;
{
	int i, j, k ;
	double tmp, tmpm ;

	if ( scaling( matrix, size ) == -1 )
		return -1 ;

	for ( k=0 ; k<size ; ++k ) {
		if ( matrix[k][k]==.0 && swap(matrix,k,size)==-1 )
			return -1 ;
		for ( j=k+1 ; j<size ; ++j ) {
			tmp = matrix[j][k]/matrix[k][k] ;
			for ( i=k+1 ; i<size+1 ; ++i )
				matrix[j][i] -= tmp*matrix[k][i] ;
		}
	}

	for ( k=size-1 ; k>=0 ; --k ) {
		for ( tmp=.0, j=k+1 ; j<size ; ++j )
			tmp += matrix[j][size] * matrix[k][j] ;
		matrix[k][size] = ( matrix[k][size] - tmp )
				/ matrix[k][k] ;
	}

	return 0 ;
}
/*
main()
{
	double **matrix ;

	matrix = (double **)malloc_2D( 4, 3, sizeof **matrix ) ;

	matrix[0][0] = 0. ;
	matrix[0][1] = 1. ;
	matrix[0][2] = 2. ;
	matrix[0][3] = 5. ;

	matrix[1][0] = 1. ;
	matrix[1][1] = 2. ;
	matrix[1][2] = 1. ;
	matrix[1][3] = 10. ;

	matrix[2][0] = 1. ;
	matrix[2][1] = -1. ;
	matrix[2][2] = 1. ;
	matrix[2][3] = 1. ;

	if ( gauss( matrix, 3 ) == -1 )
		puts( "error") ;

	printf("x1=%f, x2=%f, x3=%f\n", matrix[0][3],
		matrix[1][3], matrix[2][3] ) ;

	free_2D( 3, matrix ) ;
}
*/
