#include <stdio.h>
#include <math.h>
#include <an.h>

/*	estimate partial derivatives
	for image data stored in 1-dimensional array.

	this function assumes the following data types:
	image data: unsigned char *data
	partial derivatives: AN_VECTOR_2DTD *pd
*/
int pd_1da( data, pd, img )
	unsigned char *data ;
	AN_VECTOR_2DTD *pd ;
	AN_IMAGE_COORD img ;
{
	int x , y , t, p0l, p1l, p2l, p3l, p0h, p1h, p2h, p3h, fs ;

	fs = img.x*img.y ;

	for ( t=0; t<img.t-1; ++t ) {
		for ( y=0; y<img.y-1; ++y ) {
			for (x=0; x<img.x-1; ++x) {
				p0l = t*fs + y*img.x + x ;

				p0h = p0l+1 ;
				p1l = p0l+img.x ;
				p1h = p1l+1 ;
				p2l = p0l+fs ;
				p2h = p2l+1 ;
				p3l = p1l+fs ;
				p3h = p3l+1 ;

				pd[p0l].x = (double)
					( (int)data[p0h] - (int)data[p0l]
					+ (int)data[p1h] - (int)data[p1l]
					+ (int)data[p2h] - (int)data[p2l]
					+ (int)data[p3h] - (int)data[p3l]
					) / 4.0 ;

				p0h = p0l+img.x ;
				p1l = p0l+1 ;
				p1h = p1l+img.x ;
				p2l = p0l+fs ;
				p2h = p2l+img.x ;
				p3l = p1l+fs ;
				p3h = p3l+img.x ;

				pd[p0l].y = (double)
					( (int)data[p0h] - (int)data[p0l]
					+ (int)data[p1h] - (int)data[p1l]
					+ (int)data[p2h] - (int)data[p2l]
					+ (int)data[p3h] - (int)data[p3l]
					) / 4.0 ;

				p0l = t*fs + y*img.x + x ;
				p0h = p0l+fs ;
				p1l = p0l+1 ;
				p1h = p1l+fs ;
				p2l = p0l+img.x ;
				p2h = p2l+fs ;
				p3l = p2l+1 ;
				p3h = p3l+fs ;

				pd[p0l].t = (double)
					( (int)data[p0h] - (int)data[p0l]
					+ (int)data[p1h] - (int)data[p1l]
					+ (int)data[p2h] - (int)data[p2l]
					+ (int)data[p3h] - (int)data[p3l]
					) / 4.0 ;
			}
			p0l=img.x-2+y*img.x+t*fs ;
			p0h=p0l+1;
			pd[p0h].x = pd[p0l].x ;
			pd[p0h].y = pd[p0l].t ;
			pd[p0h].t = pd[p0l].t ;
		}

		for ( x=0; x<img.x; ++x ) {
			p0l=x+(img.y-2)*img.x+t*fs ;
			p0h=p0l+img.x ;
			pd[p0h].x = pd[p0l].x ;
			pd[p0h].y = pd[p0l].t ;
			pd[p0h].t = pd[p0l].t ;
		}
	}

	for ( x=0; x<fs; ++x ) {
		p0l=x+fs*(img.t-2) ;
		p0h=p0l+fs ;
		pd[p0h].x = pd[p0l].x ;
		pd[p0h].y = pd[p0l].t ;
		pd[p0h].t = pd[p0l].t ;
	}

	return 0 ;
}

/*********************************************************************/
/*
	partial_derivation()
*/

int high_low(simg , dhimg , dlimg , img)
	AN_IMAGE_COORD *simg , *dhimg , *dlimg , *img ;
{
	if ( simg->x > img->x - 1)
		return -1 ;
	else if (simg->x == img->x - 1)
		dhimg->x = simg->x , dlimg->x = simg->x -1 ;
	else if (simg->x < 0)
		return -1 ;
	else
		dhimg->x = simg->x + 1 , dlimg->x = simg->x ;

	if (simg->y > img->y - 1)
		return -1 ;
	else if (simg->y == img->y - 1)
		dhimg->y = simg->y , dlimg->y = simg->y - 1 ;
	else if (simg->y < 0)
		return -1 ;
	else
		dhimg->y = simg->y + 1 , dlimg->y = simg->y ;

	if (simg->t > img->t - 1)
		return -1 ;
	else if (simg->t == img->t - 1)
		dhimg->t = simg->t , dlimg->t = simg->t - 1 ;
	else if (simg->t < 0)
		return -1 ;
	else
		dhimg->t = simg->t + 1 , dlimg->t = simg->t ;

	return 0 ;
}

/*
	image data: unsigned char ***data
*/

double pdx(data , x , y , t , img)
	unsigned char ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double ex ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	ex = (double) (
	(int) data[dl.t][dl.y][dh.x] - (int) data[dl.t][dl.y][dl.x]
	+ (int) data[dl.t][dh.y][dh.x] - (int) data[dl.t][dh.y][dl.x]
	+ (int) data[dh.t][dl.y][dh.x] - (int) data[dh.t][dl.y][dl.x]
	+ (int) data[dh.t][dh.y][dh.x] - (int) data[dh.t][dh.y][dl.x]
	) / 4.0 ;

	return ex ;
}

double pdy(data , x , y , t , img)
	unsigned char ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double ey ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	ey = (double) (
	(int) data[dl.t][dh.y][dl.x] - (int) data[dl.t][dl.y][dl.x]
	+ (int) data[dl.t][dh.y][dh.x] - (int) data[dl.t][dl.y][dh.x]
	+ (int) data[dh.t][dh.y][dl.x] - (int) data[dh.t][dl.y][dl.x]
	+ (int) data[dh.t][dh.y][dh.x] - (int) data[dh.t][dl.y][dh.x]
	) / 4.0 ;

	return ey ;
}

double pdt(data , x , y , t , img)
	unsigned char ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double et ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	et = (double) (
	(int) data[dh.t][dl.y][dl.x] - (int) data[dl.t][dl.y][dl.x]
	+ (int) data[dh.t][dh.y][dl.x] - (int) data[dl.t][dh.y][dl.x]
	+ (int) data[dh.t][dl.y][dh.x] - (int) data[dl.t][dl.y][dh.x]
	+ (int) data[dh.t][dh.y][dh.x] - (int) data[dl.t][dh.y][dh.x]
	) / 4.0 ;

	return et ;
}

int pd_xyt(data , der , img)
	unsigned char ***data ;
	AN_VECTOR_2DT ***der ;
	AN_IMAGE_COORD img ;
{
	int x , y , t ;

	for (t = 0 ; t < img.t ; ++t)
		for (y = 0 ; y < img.y ; ++y)
			for (x = 0 ; x < img.x ; ++x) {
				der[t][y][x].x =
					pdx(data, x, y, t, img) ;
				der[t][y][x].y =
					pdy(data, x, y, t, img) ;
				der[t][y][x].t =
					pdt(data, x, y, t, img) ;
			}

	return 0 ;
}

/*
	partial_derivation2()
*/

double pdx2( data, x, y, t, img)
	unsigned char ***data ;
	int x, y, t ;
	AN_IMAGE_COORD img ;
{
	double ex ;

	if (x == 0 || y == 0 || t== 0 ||
		x == img.x - 1 || y == img.y - 1 || t == img.t - 1)
		ex = pdx( data, x, y, t, img) ;
	else
		ex = (double) (
		  (int)data[t-1][y-1][x+1]-(int)data[t-1][y-1][x-1]
		+ (int)data[t-1][y][x+1]-(int)data[t-1][y][x-1]
		+ (int)data[t-1][y+1][x+1]-(int)data[t-1][y+1][x-1]
		+ (int)data[t][y-1][x+1]-(int)data[t][y-1][x-1]
		+ (int)data[t][y][x+1]-(int)data[t][y][x-1]
		+ (int)data[t][y+1][x+1]-(int)data[t][y+1][x-1]
		+ (int)data[t+1][y-1][x+1]-(int)data[t+1][y-1][x-1]
		+ (int)data[t+1][y][x+1]-(int)data[t+1][y][x+1]
		+ (int)data[t+1][y+1][x+1]-(int)data[t+1][y+1][x-1]
		) / 2.0 / 9.0 ;

	return ex ;
}

double pdy2(data , x , y , t , img)
	unsigned char ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double ey ;

	if (x == 0 || y == 0 || t== 0 ||
		x == img.x - 1 || y == img.y - 1 || t == img.t - 1)
		ey = pdy( data, x, y, t, img ) ;
	else
		ey = (double) (
		  (int)data[t-1][y+1][x-1]-(int)data[t-1][y-1][x-1]
		+ (int)data[t-1][y+1][x]-(int)data[t-1][y-1][x]
		+ (int)data[t-1][y+1][x+1]-(int)data[t-1][y-1][x+1]
		+ (int)data[t][y+1][x-1]-(int)data[t][y-1][x-1]
		+ (int)data[t][y+1][x]-(int)data[t][y-1][x]
		+ (int)data[t][y+1][x+1]-(int)data[t][y-1][x+1]
		+ (int)data[t+1][y+1][x-1]-(int)data[t+1][y-1][x-1]
		+ (int)data[t+1][y+1][x]-(int)data[t+1][y-1][x]
		+ (int)data[t+1][y+1][x+1]-(int)data[t+1][y-1][x+1]
		) / 2.0 / 9.0 ;

	return ey ;
}

double pdt2(data , x , y , t , img)
	unsigned char ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double et ;

	if (x == 0 || y == 0 || t== 0 ||
		x == img.x - 1 || y == img.y - 1 || t == img.t - 1)
		et = pdt( data, x, y, t, img) ;
	else
		et = (double) (
		(int)data[t+1][y-1][x-1]-(int)data[t-1][y-1][x-1]
		+ (int)data[t+1][y-1][x]-(int)data[t-1][y-1][x]
		+ (int)data[t+1][y-1][x+1]-(int)data[t-1][y-1][x+1]
		+ (int)data[t+1][y][x-1]-(int)data[t-1][y][x-1]
		+ (int)data[t+1][y][x]-(int)data[t-1][y][x]
		+ (int)data[t+1][y][x+1]-(int)data[t-1][y][x+1]
		+ (int)data[t+1][y+1][x-1]-(int)data[t-1][y+1][x-1]
		+ (int)data[t+1][y+1][x]-(int)data[t-1][y+1][x]
		+ (int)data[t+1][y+1][x+1]-(int)data[t-1][y+1][x+1]
		) / 2.0 / 9.0 ;

	return et ;
}

int pd_xyt2(data , der , img)
	unsigned char ***data ;
	AN_VECTOR_2DT ***der ;
	AN_IMAGE_COORD img ;
{
	int x , y , t ;

	for ( t=0 ; t<img.t ; ++t)
		for ( y=0 ; y<img.y ; ++y)
			for ( x=0 ; x<img.x ; ++x) {
				der[t][y][x].x =
					pdx2( data, x, y, t, img ) ;
				der[t][y][x].y =
					pdy2( data, x, y, t, img ) ;
				der[t][y][x].t =
					pdt2( data, x, y, t, img ) ;
			}

	return 0 ;
}

/*	image data: float	*/
double fpdx(data , x , y , t , img)
	float ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double ex ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	ex = (
	  data[dl.t][dl.y][dh.x] - data[dl.t][dl.y][dl.x]
	+ data[dl.t][dh.y][dh.x] - data[dl.t][dh.y][dl.x]
	+ data[dh.t][dl.y][dh.x] - data[dh.t][dl.y][dl.x]
	+ data[dh.t][dh.y][dh.x] - data[dh.t][dh.y][dl.x]
	) / 4.0 ;

	return ex ;
}

double fpdy(data , x , y , t , img)
	float ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double ey ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	ey = (
	  data[dl.t][dh.y][dl.x] - data[dl.t][dl.y][dl.x]
	+ data[dl.t][dh.y][dh.x] - data[dl.t][dl.y][dh.x]
	+ data[dh.t][dh.y][dl.x] - data[dh.t][dl.y][dl.x]
	+ data[dh.t][dh.y][dh.x] - data[dh.t][dl.y][dh.x]
	) / 4.0 ;

	return ey ;
}

double fpdt(data , x , y , t , img)
	float ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double et ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	et = (
	data[dh.t][dl.y][dl.x] - data[dl.t][dl.y][dl.x]
	+ data[dh.t][dh.y][dl.x] - data[dl.t][dh.y][dl.x]
	+ data[dh.t][dl.y][dh.x] - data[dl.t][dl.y][dh.x]
	+ data[dh.t][dh.y][dh.x] - data[dl.t][dh.y][dh.x]
	) / 4.0 ;

	return et ;
}

/*	image data: float	*/
double fpdx2( data, x, y, t, img)
	float ***data ;
	int x, y, t ;
	AN_IMAGE_COORD img ;
{
	double ex ;

	if (x == 0 || y == 0 || t== 0 ||
		x == img.x - 1 || y == img.y - 1 || t == img.t - 1)
		ex = pdx( data, x, y, t, img) ;
	else
		ex = (
		  data[t-1][y-1][x+1]-data[t-1][y-1][x-1]
		+ data[t-1][y][x+1]-data[t-1][y][x-1]
		+ data[t-1][y+1][x+1]-data[t-1][y+1][x-1]
		+ data[t][y-1][x+1]-data[t][y-1][x-1]
		+ data[t][y][x+1]-data[t][y][x-1]
		+ data[t][y+1][x+1]-data[t][y+1][x-1]
		+ data[t+1][y-1][x+1]-data[t+1][y-1][x-1]
		+ data[t+1][y][x+1]-data[t+1][y][x+1]
		+ data[t+1][y+1][x+1]-data[t+1][y+1][x-1]
		) / 2.0 / 9.0 ;

	return ex ;
}

double fpdy2(data , x , y , t , img)
	float ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double ey ;

	if (x == 0 || y == 0 || t== 0 ||
		x == img.x - 1 || y == img.y - 1 || t == img.t - 1)
		ey = pdy( data, x, y, t, img ) ;
	else
		ey = (
		  data[t-1][y+1][x-1]-data[t-1][y-1][x-1]
		+ data[t-1][y+1][x]-data[t-1][y-1][x]
		+ data[t-1][y+1][x+1]-data[t-1][y-1][x+1]
		+ data[t][y+1][x-1]-data[t][y-1][x-1]
		+ data[t][y+1][x]-data[t][y-1][x]
		+ data[t][y+1][x+1]-data[t][y-1][x+1]
		+ data[t+1][y+1][x-1]-data[t+1][y-1][x-1]
		+ data[t+1][y+1][x]-data[t+1][y-1][x]
		+ data[t+1][y+1][x+1]-data[t+1][y-1][x+1]
		) / 2.0 / 9.0 ;

	return ey ;
}

double fpdt2(data , x , y , t , img)
	float ***data ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double et ;

	if (x == 0 || y == 0 || t== 0 ||
		x == img.x - 1 || y == img.y - 1 || t == img.t - 1)
		et = pdt( data, x, y, t, img) ;
	else
		et = (
		data[t+1][y-1][x-1]-data[t-1][y-1][x-1]
		+ data[t+1][y-1][x]-data[t-1][y-1][x]
		+ data[t+1][y-1][x+1]-data[t-1][y-1][x+1]
		+ data[t+1][y][x-1]-data[t-1][y][x-1]
		+ data[t+1][y][x]-data[t-1][y][x]
		+ data[t+1][y][x+1]-data[t-1][y][x+1]
		+ data[t+1][y+1][x-1]-data[t-1][y+1][x-1]
		+ data[t+1][y+1][x]-data[t-1][y+1][x]
		+ data[t+1][y+1][x+1]-data[t-1][y+1][x+1]
		) / 2.0 / 9.0 ;

	return et ;
}


/*	partial derivation for a vector field */

int pdvx( v, pd , x , y , t , img )
	AN_VECTOR_FF ***v ;
	AN_VECTOR_FF *pd ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	double ex ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if ( high_low(&simg , &dh , &dl , &img) == -1 ) {
		(*pd).vx = (float).0 ;
		(*pd).vy = (float).0 ;
		return -1 ;
	} else {
		(*pd).vx = (float) (
			( v[dl.t][dl.y][dh.x].vx - v[dl.t][dl.y][dl.x].vx
			+ v[dl.t][dh.y][dh.x].vx - v[dl.t][dh.y][dl.x].vx
			+ v[dh.t][dl.y][dh.x].vx - v[dh.t][dl.y][dl.x].vx
			+ v[dh.t][dh.y][dh.x].vx - v[dh.t][dh.y][dl.x].vx
			) / 4.0 ) ;
		(*pd).vy = (float) (
			( v[dl.t][dl.y][dh.x].vy - v[dl.t][dl.y][dl.x].vy
			+ v[dl.t][dh.y][dh.x].vy - v[dl.t][dh.y][dl.x].vy
			+ v[dh.t][dl.y][dh.x].vy - v[dh.t][dl.y][dl.x].vy
			+ v[dh.t][dh.y][dh.x].vy - v[dh.t][dh.y][dl.x].vy
			) / 4.0 ) ;
		return 0 ;
	}
}

int pdvy( v, pd , x , y , t , img )
	AN_VECTOR_FF ***v ;
	AN_VECTOR_FF *pd ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if ( high_low(&simg , &dh , &dl , &img) == -1 ) {
		(*pd).vx = (float).0 ;
		(*pd).vy = (float).0 ;
		return -1 ;
	} else {
		(*pd).vx = (float) (
			( v[dl.t][dh.y][dl.x].vx - v[dl.t][dl.y][dl.x].vx
			+ v[dl.t][dh.y][dh.x].vx - v[dl.t][dl.y][dh.x].vx
			+ v[dh.t][dh.y][dl.x].vx - v[dh.t][dl.y][dl.x].vx
			+ v[dh.t][dh.y][dh.x].vx - v[dh.t][dl.y][dh.x].vx
			) / 4.0 ) ;
		(*pd).vy = (float) (
			( v[dl.t][dh.y][dl.x].vy - v[dl.t][dl.y][dl.x].vy
			+ v[dl.t][dh.y][dh.x].vy - v[dl.t][dl.y][dh.x].vy
			+ v[dh.t][dh.y][dl.x].vy - v[dh.t][dl.y][dl.x].vy
			+ v[dh.t][dh.y][dh.x].vy - v[dh.t][dl.y][dh.x].vy
			) / 4.0 ) ;
		return 0 ;
	}
}

int pdvt( v, pd , x , y , t , img )
	AN_VECTOR_FF ***v ;
	AN_VECTOR_FF *pd ;
	int x , y , t ;
	AN_IMAGE_COORD img ;
{
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y , simg.t = t ;
	if ( high_low(&simg , &dh , &dl , &img) == -1 ) {
		(*pd).vx = (float).0 ;
		(*pd).vy = (float).0 ;
		return -1 ;
	} else {
		(*pd).vx = (float) (
			( v[dh.t][dl.y][dl.x].vx - v[dl.t][dl.y][dl.x].vx
			+ v[dh.t][dl.y][dh.x].vx - v[dl.t][dl.y][dh.x].vx
			+ v[dh.t][dh.y][dl.x].vx - v[dl.t][dh.y][dl.x].vx
			+ v[dh.t][dh.y][dh.x].vx - v[dl.t][dh.y][dh.x].vx
			) / 4.0 ) ;
		(*pd).vy = (float) (
			( v[dh.t][dl.y][dl.x].vy - v[dl.t][dl.y][dl.x].vy
			+ v[dh.t][dl.y][dh.x].vy - v[dl.t][dl.y][dh.x].vy
			+ v[dh.t][dh.y][dl.x].vy - v[dl.t][dh.y][dl.x].vy
			+ v[dh.t][dh.y][dh.x].vy - v[dl.t][dh.y][dh.x].vy
			) / 4.0 ) ;
		return 0 ;
	}
}

/*	for 2-dimensional partial derivations	*/

double pdx2d( data , x , y , img)
	unsigned char **data ;
	int x , y ;
	AN_IMAGE_COORD img ;
{
	double ex ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y, simg.t=0 ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	ex = (double) (
		(int) data[dl.y][dh.x] - (int) data[dl.y][dl.x]
		+ (int) data[dh.y][dh.x] - (int) data[dh.y][dl.x]
		) / 2.0 ;

	return ex ;
}

double pdy2d( data , x , y , img)
	unsigned char **data ;
	int x , y ;
	AN_IMAGE_COORD img ;
{
	double ey ;
	AN_IMAGE_COORD simg , dh , dl ;

	simg.x = x , simg.y = y, simg.t = 0 ;
	if (high_low(&simg , &dh , &dl , &img) == -1)
		return .0 ;

	ey = (double) (
		(int) data[dh.y][dl.x] - (int) data[dl.y][dl.x]
		+ (int) data[dh.y][dh.x] - (int) data[dl.y][dh.x]
		) / 2.0 ;

	return ey ;
}

double pdx2d2( data, x, y, img)
	unsigned char **data ;
	int x, y ;
	AN_IMAGE_COORD img ;
{
	double ex ;

	if (x == 0 || y == 0 ||
		x == img.x - 1 || y == img.y - 1 )
		ex = pdx2d( data, x, y, img) ;
	else
		ex = (double) (
		  (int)data[y-1][x+1]	-(int)data[y-1][x-1]
		+ (int)data[y][x+1]	-(int)data[y][x-1]
		+ (int)data[y+1][x+1]	-(int)data[y+1][x-1]
		) / 2.0 / 3.0 ;

	return ex ;
}

double pdy2d2(data , x , y , img)
	unsigned char **data ;
	int x , y ;
	AN_IMAGE_COORD img ;
{
	double ey ;

	if (x == 0 || y == 0 ||
		x == img.x - 1 || y == img.y - 1 )
		ey = pdy2d( data, x, y, img ) ;
	else
		ey = (double) (
		  (int)data[y+1][x-1]	-(int)data[y-1][x-1]
		+ (int)data[y+1][x]	-(int)data[y-1][x]
		+ (int)data[y+1][x+1]	-(int)data[y-1][x+1]
		) / 2.0 / 3.0 ;

	return ey ;
}

double laplacian( data, x, y, t, img )
	unsigned char ***data ;
	int x, y, t ;
	AN_IMAGE_COORD img ;
{
	int lt, ht ;
	double lap ;

	if ( x<0 || x>img.x-1 || y<0 || y>img.y-1 || t<0 || t>img.t-1 )
		return .0 ;

	if ( x==0 ) x=1 ;
	if ( x==img.x-1 ) x=img.x-2 ;

	if ( y==0 ) y=1 ;
	if ( y==img.y-1 ) y=img.y-2 ;

	lt = ( (t==0) ? 0: t-1) ;
	ht = ( (t==img.t-1) ? img.t-1: t+1 ) ;

	for ( lap=.0, t=lt; t<=ht; ++t )
/*
		lap +=	( (double)data[t][y][x+1]/6.
			+ (double)data[t][y][x-1]/6.
			+ (double)data[t][y+1][x]/6.
			+ (double)data[t][y-1][x]/6.
			+ (double)data[t][y+1][x+1]/12.
			+ (double)data[t][y+1][x-1]/12.
			+ (double)data[t][y-1][x+1]/12.
			+ (double)data[t][y-1][x-1]/12.
			- (double)data[t][y][x] ) ;

	return (lap/(double)(ht-lt+1)) ;
*/
		lap += ( (double)data[t][y][x-1]
			+ (double)data[t][y][x+1]
			+ (double)data[t][y-1][x]
			+ (double)data[t][y+1][x]
			- (double)data[t][y][x] * 4. ) ;

	return (lap/(double)(ht-lt+1)) ;
}

double loglap( data, x, y, t, img )
	unsigned char ***data ;
	int x, y, t ;
	AN_IMAGE_COORD img ;
{
	int lt, ht ;
	double lap ;

	if ( x<0 || x>img.x-1 || y<0 || y>img.y-1 || t<0 || t>img.t-1 )
		return .0 ;

	if ( x==0 ) x=1 ;
	if ( x==img.x-1 ) x=img.x-2 ;

	if ( y==0 ) y=1 ;
	if ( y==img.y-1 ) y=img.y-2 ;

	lt = ( (t==0) ? 0: t-1) ;
	ht = ( (t==img.t-1) ? img.t-1: t+1 ) ;

	for ( lap=.0, t=lt; t<=ht; ++t )
/*
		lap +=	( (double)data[t][y][x+1]/6.
			+ (double)data[t][y][x-1]/6.
			+ (double)data[t][y+1][x]/6.
			+ (double)data[t][y-1][x]/6.
			+ (double)data[t][y+1][x+1]/12.
			+ (double)data[t][y+1][x-1]/12.
			+ (double)data[t][y-1][x+1]/12.
			+ (double)data[t][y-1][x-1]/12.
			- (double)data[t][y][x] ) ;

	return (lap/(double)(ht-lt+1)) ;
*/
		lap += ( log( (double)data[t][y][x-1] )
			+ log( (double)data[t][y][x+1] )
			+ log( (double)data[t][y-1][x] )
			+ log( (double)data[t][y+1][x] )
			- log( (double)data[t][y][x] ) * 4. ) ;

	return (lap/(double)(ht-lt+1)) ;
}

double ip2d( data, x, y, img )
	unsigned char **data ;
	double x, y ;
	AN_IMAGE_COORD img ;
{
	double d ;
	double lx, ly ;
	int x0, y0 ;

	if ( x<.0 || x>(double)(img.x-2) ||
		y<.0 || y>(double)(img.y-2) )
		return .0 ;

	x0 = (int)floor(x) ;
	y0 = (int)floor(y) ;
	lx = x-(double)x0 ;
	ly = y-(double)y0 ;

	d =	(1.-lx)*(1.-ly)*(double)data[y0][x0]
		+ (1.-lx)*ly*(double)data[y0+1][x0]
		+ lx*(1.-ly)*(double)data[y0][x0+1]
		+ lx*ly*(double)data[y0+1][x0+1] ;

	return d ;
}

